#pragma once

#include <string>
#include <vector>
#include <memory>
#include <unordered_map>

#include "tf_data_types.h"

namespace Trueface
{
    class SDK {
        public:
            /**
             * Initialize the SDK using default \ref ConfigurationOptions.
             *
             * @sa ConfigurationOptions
             */
            SDK();

            /**
             * Initialize the SDK using custom \ref ConfigurationOptions.
             *
             * @param options custom configuration options.
             */
            SDK(const ConfigurationOptions& options);

            ~SDK();

            /**
             *  Sets and validates the given license token. Need to call this method
             *  before being able to use the SDK.
             *
             * @param[in] token the license token (if you do not have this talk to
             *        support@trueface.ai).
             * @return Whether the given license token is valid.
             * 
             * @sa isLicensed()
            */
            bool setLicense(std::string token);
#if defined(TARGET_OS_IPHONE) || defined(TARGET_IPHONE_SIMULATOR)
            bool setLicense(std::string token, std::string package_id);
#endif
            /** 
             * Checks whether the given license token is valid and you can use
             * the SDK.
             *
             * @param[out] daysRemaining the number of remaining days that the token is valid for.
             * @return Whether the given license token is valid.
             * 
             * @sa setLicense()
            */
            bool isLicensed(unsigned int& daysRemaining);

            /** 
             * Gets the version-build number of the SDK.
             *
             *
             * @return Version Number as a string.
            */
            static std::string getVersion();

            /**
             *  Set the image that is processed by the other methods.
             *
             * @param[in] image an 8-bit decoded image array, in the CPU memory or the GPU memory.
             * @param[in] width the image width.
             * @param[in] height the image height.
             * @param[in] color the image color model, see ColorCode.
             * @param[in] stride the distance between image array rows in bytes, also known as step size (used only for pointers to the GPU memory).
             *
             * @return error code, see ErrorCode.
             * Note, it is highly encouraged to check the return value from setImage before proceeding.
             * If the license is invalid, the INVALID_LICENSE error will be returned.
             *
             * @sa setImage(std::string imageFile)
            */
            ErrorCode setImage(uint8_t* image, uint16_t width, uint16_t height,
                               ColorCode color, int stride=0);


            /**
             *  Set the image that is processed by the other methods.
             *
             * @param[in] encodedImageBuffer encoded image buffer, supports JPG, PNG, BMP, and TIFF.
             *
             * @return error code, see ErrorCode.
             * Note, it is highly encouraged to check the return value from setImage before proceeding.
             * If the license is invalid, the INVALID_LICENSE error will be returned.
             *
             * @sa setImage(std::string imageFile)
            */
            ErrorCode setImage(const std::vector<uint8_t>& encodedImageBuffer);

            /** 
             * Set the image that is processed by the other methods.
             *
             * @param[in] imageFile the path of a JPEG, JPG, PNG, BMP, or TIFF file (path must not contain ~ ).
             *
             * @return error code, see ErrorCode.
             * Note, it is highly encouraged to check the return value from setImage before proceeding.
             * If the license is invalid, the INVALID_LICENSE error will be returned.
             * 
             * @sa setImage(uint8_t* image, uint16_t width, uint16_t height,
             *  ColorCode color, int stride)
            */

            ErrorCode setImage(const std::string& imageFile);

            /** 
             * Detect people and objects in the image.
             *
             * @param[out] boundingBoxes a vector of BoundingBox objects that
             *             gets populated with the detected objects.
             * 
             * @sa getObjectLabelString(ObjectLabel label)
            */
            void detectObjects(std::vector<BoundingBox>& boundingBoxes);

            /** 
             * Estimate body pose.
             *
             * @param[in] boundingBoxes bounding boxes returned by detectObjects()
             * @param[out] landmarks a vector of Landmark objects that
             *             represent body joints.
            */
            void estimatePose(const std::vector<BoundingBox>& boundingBoxes, std::vector<std::vector<Landmark>>& landmarks);

            /** 
             * Draw body pose.
             *
             * @param[in] filepath relative or absolute file path without a file extension.
             * @param[in] landmarks a vector of Landmark objects that
             *             represent body joints.
            */
            void drawPose(const std::string& filepath, const std::vector<std::vector<Landmark>>& landmarks);

            /**
             *  Converts an object label to a string. This can be used to
             *  display the object labels.
             * 
             * @param[in] label the label of a BoundingBox.
             * @return The object label as a string.
            */
            static std::string getObjectLabelString(ObjectLabel label);

            /**
             * Detect the largest face in the image.
             * This method has a small false positive rate.
             * To reduce the false positive rate to near zero, filter out faces with score lower than 0.90.
             * Alternatively, you can use the `Trueface::FaceDetectionFilter`
             * configuration option to filter the detected faces.
             * See detectFaces() for the detection scale range.
             *
             * @param[out] faceBoxAndLandmarks the FaceBoxAndLandmarks containing the landmarks and bounding box
             * of the largest detected face in the image.
             * @param[out] found whether a face was found in the image.
             *
             * @return error code, see ErrorCode.
            */
            ErrorCode detectLargestFace(FaceBoxAndLandmarks& faceBoxAndLandmarks,
                                   bool &found);

            /**
             * Detect all the faces in the image. This method has a small false positive rate.
             * To reduce the false positive rate to near zero, filter out faces with score lower than 0.90.
             * Alternatively, you can use the `Trueface::FaceDetectionFilter`
             * configuration option to filter the detected faces.
             *
             * The face detector has a detection scale range of about 5 octaves. \ref ConfigurationOptions.smallestFaceHeight
             * determines the lower of the detection scale range. E.g., setting \ref ConfigurationOptions.smallestFaceHeight to
             * 40 pixels yields the detection scale range of ~40 pixels to 1280 (=40x2^5) pixels.
             *
             * @param[out] faceBoxAndLandmarks a vector of \ref FaceBoxAndLandmarks representing each of the detected faces.
             * If not faces are found, the vector will be empty. The detected faces are sorted in order of descending face score.
             *
             * @return error code, see ErrorCode.
             */
            ErrorCode detectFaces(std::vector<FaceBoxAndLandmarks>& faceBoxAndLandmarks);

            /**
             * Obtain the 106 face landmarks.
             *
             * @param[in] faceBoxAndLandmarks FaceBoxAndLandmarks returned by detectFaces() or detectLargestFace().
             * @param[out] points a vector of 106 face landmark points.
             *
             * @return error code, see ErrorCode.
            */
            [[deprecated("Deprecated due to suboptimal performance. Use detectFaces instead.")]]
            ErrorCode getFaceLandmarks(const FaceBoxAndLandmarks& faceBoxAndLandmarks, std::vector<Point<int>>& points);

            /**
             *  Align the the detected face to be optimized for passing to
             *  feature extraction. If using the face chip with Trueface algorithms (ex face recognition),
             *  do not change the default margin and scale values.
             *
             * @param[in] faceBoxAndLandmarks the FaceBoxAndLandmarks returned
             *            by detectLargestFace() or detectFaces().
             * @param[out] faceImage the pointer to a uint8_t buffer of 
             *             112x112x3 = 37632 bytes (when using default margins and scale).
             *             The aligned face image is stored in this buffer.
             *             The memory must be allocated by the user.
             *             If using non-default margin and scale
             *             (again, non-standard face chip sizes will not work with Trueface algorithms),
             *             the faceImage will be of size:
             *             width = int((112+marginLeft+marginRight)*scale),
             *             height = int((112+marginTop+marginBottom)*scale),
             *             and therefore the buffer size is computed as:
             *             width * height * 3
             *
             * @param[in] marginLeft adds a margin to the left side of the face chip.
             * @param[in] marginTop adds a margin to the top side of the face chip.
             * @param[in] marginRight adds a margin to the right side of the face chip.
             * @param[in] marginBottom adds a margin to the bottom side of the face chip.
             * @param[in] scale changes the scale of the face chip.
             *
             * @return error code, see ErrorCode.
            */
            ErrorCode extractAlignedFace(const FaceBoxAndLandmarks& faceBoxAndLandmarks,
                                         uint8_t* faceImage,
                                         int marginLeft=0,
                                         int marginTop=0,
                                         int marginRight=0,
                                         int marginBottom=0,
                                         float scale=1.0);

            /**
             * Store the face image in JPEG file.
             *
             * @param[in] faceImage the extracted image by extractAlignedFace().
             * Face image must have size of 112x112 pixels (default extractAlignedFace() margin and scale values).
             * @param[in] filepath relative or absolute file path without a file extension.
            */
            void saveFaceImage(uint8_t* faceImage, const std::string& filepath);

            /**
             * Detect the largest face in the image and return its feature
             * vector.
             *
             * @param[out] faceprint a Faceprint object which will contain
             *             the face feature vector.
             * @param[out] foundFace indicates if a face was detected in the image.
             * If no face was detected, then the faceprint will be empty.
             *
             * @return error code, see ErrorCode.
            */
            ErrorCode getLargestFaceFeatureVector(Faceprint &faceprint, bool& foundFace);

            /**
             * Extract the face feature vector from an aligned face image.
             *
             * @param[in] alignedFaceImage buffer returned by extractAlignedFace().
             * Face image must be have size of 112x112 pixels (default extractAlignedFace() margin and scale values).
             * @param[out] faceprint a Faceprint object which will contain
             *             the face feature vector.
             *
             * @return error code, see ErrorCode.
            */
            ErrorCode getFaceFeatureVector(uint8_t* alignedFaceImage, Faceprint& faceprint);

            /**
             * Extract the face feature vectors from the aligned face images.
             * This batch processing method increases throughput on platforms such as cuda.
             *
             * @param[in] alignedFaceImages vector of aligned face image buffers.
             * @param[out] faceprints vector of Faceprint object which will contain
             *             the face feature vectors.
             *
             * @return error code, see ErrorCode.
            */
            ErrorCode getFaceFeatureVectors(std::vector<uint8_t*>& alignedFaceImages, std::vector<Faceprint>& faceprints);

            /**
             * Extract the face feature vector from the face box.
             *
             * @param[in] faceBoxAndLandmarks the FaceBoxAndLandmarks returned by detectFaces() or detectLargestFace().
             * @param[out] faceprint Faceprint to contain the face template
             *
             * @return error code, see ErrorCode.
            */
            ErrorCode getFaceFeatureVector(const FaceBoxAndLandmarks& faceBoxAndLandmarks, Faceprint& faceprint);

            /**
             * Convert a Faceprint into a json string.
             *
             * @param[in] faceprint The Faceprint to stringify.
             *
             * @return The json string representation of the Faceprint.
            */
            static std::string faceprintToJson(const Faceprint& faceprint);

            /**
             * Populate a Faceprint from a json string.
             *
             * @param[in] jsonStr The json string representation of a Faceprint, generated from the faceprintToJson() function.
             * @param[out] faceprint The Faceprint which will be populated from the json string.
             *
             * @return error code, see ErrorCode.
            */
            static ErrorCode jsonToFaceprint(const std::string& jsonStr, Faceprint& faceprint);

            /**
             * Compute the similarity between two feature vectors, or how
             * similar two faces are.
             *
             * @param[in] faceprint1 the first Faceprint to be compared.
             * @param[in] faceprint2 the second Faceprint to be compared.
             * @param[out] matchProbability the probability the two face feature vectors are a match.
             * @param[out] similarityMeasure the computed similarity measure.
             *
             * @return error code, see ErrorCode.
             *
             * @sa getLargestFaceFeatureVector
            */
            static ErrorCode getSimilarity(const Faceprint& faceprint1,
                                    const Faceprint& faceprint2,
                                    float &matchProbability,
                                    float &similarityMeasure);

            /**
             * Estimate the head pose.
             *
             * @param[in] faceBoxAndLandmarks FaceBoxAndLandmarks returned by detectFaces() or detectLargestFace().
             * @param[out] yaw the rotation angle around the image's vertical axis, in radians.
             * @param[out] pitch the rotation angle around the image's transverse axis, in radians.
             * @param[out] roll the rotation angle around the image's longitudinal axis, in radians.
             *
             * @return error code, see ErrorCode.
            */
            ErrorCode estimateHeadOrientation(const FaceBoxAndLandmarks& faceBoxAndLandmarks, float& yaw, float& pitch, float&roll);

            /**
             * Estimate the quality of the face image for recognition.
             *
             * @param[in] alignedFaceImage The array returned by extractAlignedFace().
             * @param[out] quality a value between 0 to 1, 1 being prefect quality for recognition.
             * We suggest using a threshold of 0.999 as a filter for enrollment images.
             *
             * @return error code, see ErrorCode.
            */
            ErrorCode estimateFaceImageQuality(uint8_t* alignedFaceImage, float& quality);

            /**
             * Estimate score for eye blink.
             *
             * @param[out] leftEyeWinkScore score indicating the left eye is open (0 is closed, 1 is open).
             * @param[out] rightEyeWinkScore score indicating the right eye is open (0 is closed, 1 is open).
             *
             * Optimal score threshold for distinguishing between open and closed is 0.4 in normal use and 0.1
             * in low light environments.
             *
             * @return error code, see ErrorCode.
            */
            [[deprecated("Deprecated due to suboptimal performance")]]
            ErrorCode detectBlink(float &leftEyeWinkScore, float &rightEyeWinkScore);

            /**
             * Detect if there is a presentation attack attempt.
             *
             * @param[in] faceBoxAndLandmarks FaceBoxAndLandmarks returned by detectFaces() or detectLargestFace().
             * @param[out] result the SpoofLabel results for the face image.
             * @param[out] spoofScore the probability that the image is a spoof attempt
             * (1 indicates real image, 0 indicates spoof attempt).
             * @param[out] threshold the spoof score threshold above which it is considered a spoof attempt.
             *
             * @return error code, see ErrorCode.
            */
            [[deprecated("Deprecated due to suboptimal performance")]]
            ErrorCode detectSpoof(const FaceBoxAndLandmarks& faceBoxAndLandmarks, SpoofLabel& result, float& spoofScore, float threshold = 0.5);

            /**
             *
             * Create a connection to a new or existing database.
             * If the database does not exist, a new one will be created with the provided name.
             * If the `Trueface::DatabaseManagementSystem::NONE` (memory only)
             * configuration option is selected, this function does not need to be called (and is a harmless no-op).
             *
             * @param databaseConnectionString
             * If `Trueface::DatabaseManagementSystem::SQLITE` is selected, this should be the filepath to the database.
             * ex. "/myPath/myDatabase.db".
             * If `Trueface::DatabaseManagementSystem::POSTGRESQL`
             * is selected, this should be a database connection string.
             * <a href="https://www.postgresql.org/docs/current/libpq-connect.html#LIBPQ-PARAMKEYWORDS">Here</a>
             * is a list of all supported PostgreSQL connection parameters.
             * ex. "hostaddr=192.168.1.0 port=5432 dbname=face_recognition user=postgres password=my_password".
             * ex. "host=localhost port=5432 dbname=face_recognition user=postgres password=m_password".
             * To enable ssl, add "sslmode=require" to the connection string.
             *
             * @return error code, see ErrorCode.
             */

            ErrorCode createDatabaseConnection(const std::string& databaseConnectionString);

            /**
             *
             * Create a new collection, or load data from an existing collection into memory (RAM) if one with the
             * provided name already exists in the database.
             * Equivalent to calling createCollection() then loadCollection().
             *
             * @param[in] collectionName the name of the collection.
             *
             * @return error code, see ErrorCode.
             */
            ErrorCode createLoadCollection(const std::string& collectionName);

            /**
             *
             * Create a new collection in the database.
             * Before enrolling Faceprints into newly created collection, must call loadCollection().
             * If the collection with the provided name already exists, this is a harmless no-op.
             *
             * @param[in] collectionName the name of the collection.
             *
             * @return error code, see ErrorCode.
             */
            ErrorCode createCollection(const std::string& collectionName);

            /**
             *
             * Loads the collection into memory.
             * Must be called before enrolling Faceprints or calling identification functions.
             *
             * @param[in] collectionName the name of the collection to load into memory.
             *
             * @return error code, see ErrorCode.
             * Will return an error if the collection does not exist.
             */
            ErrorCode loadCollection(const std::string& collectionName);

            /**
             *
             * Deletes a collection from the current database.
             *
             * @param[in] collectionName the name of the collection to delete.
             *
             * @return error code, see ErrorCode.
             * Will return an error if the collection does not exist.
             */
            ErrorCode deleteCollection(const std::string& collectionName);

            /**
             * Enroll a Faceprint for a new or existing identity in the collection.
             *
             * @param[in] faceprint the Faceprint to enroll in the collection.
             * @param[in] identity the identity corresponding to the Faceprint.
             * @param[out] UUID universally unique identifier corresponding to the Faceprint.
             *
             * @return error code, see ErrorCode.
             */
            ErrorCode enrollFaceprint(const Faceprint& faceprint, const std::string& identity, std::string& UUID);

            /**
             * Get a list of the names of all the collections in the database.
             * Collection names can then be passed to getCollectionMetadata() and getCollectionIdentities().
             *
             * @param[out] collectionNames list of collection names in the database.
             *
             * @return error code, see ErrorCode.
             */
            ErrorCode getCollectionNames(std::vector<std::string>& collectionNames);

            /**
             * Get the metadata for the specified collection in the database, loaded or unloaded.
             *
             * @param[in] collectionName the name of the collection for which to retrieve the metadata.
             * @param[out] metadata the metadata for the specified collection.
             *
             * @return error code, see ErrorCode.
             */
            ErrorCode getCollectionMetadata(const std::string& collectionName, CollectionMetadata& metadata);

            /**
             * Get a map of identities and UUIDs for the specified collection in the database, loaded or unloaded.
             * This can be a slow operation (especially for unloaded collections), call sparingly.
             *
             * @param[in] collectionName the name of the collection for which to retrieve the identities.
             * @param[out] identities a map of identities and corresponding UUIDs.
             * There can be multiple UUIDs mapped to a single identity if multiple Faceprints were enrolled for that identity.
             *
             * @return error code, see ErrorCode.
             */
            ErrorCode getCollectionIdentities(const std::string& collectionName,
                                                  std::unordered_multimap<std::string, std::string>& identities);

            /**
             * Remove a Faceprint from the collection using the UUID.
             *
             * @param[in] UUID the universally unique identifier returned by enrollFaceprint().
             *
             * @return error code, see ErrorCode.
             */
            ErrorCode removeByUUID(const std::string& UUID);

            /**
             * Remove all Faceprints in the collection corresponding to the identity.
             *
             * @param[in] identity the identity to remove from the collection.
             * @param[out] numFaceprintsRemoved the the number of Faceprint which were removed for that identity.
             *
             * @return error code, see ErrorCode.
             */
            ErrorCode removeByIdentity(const std::string& identity, unsigned int& numFaceprintsRemoved);

            /**
             * Get the top match Candidate in the collection and the corresponding similarity score and match probability.
             *
             * @param[in] faceprint the Faceprint to be identified.
             * @param[out] candidate the top match Candidate.
             * @param[out] found set to true if a match is found.
             * @param[in] threshold the similarity score threshold above which it is considered a match.
             * Higher thresholds may result in faster queries.
             * Refer to our <a href="https://docs.trueface.ai/ROC-Curves-d47d2730cf0a44afacb39aae0ed1b45a">ROC curves</a> when selecting a threshold.
             *
             * @return error code, see ErrorCode.
             */
            ErrorCode identifyTopCandidate(const Faceprint& faceprint, Candidate& candidate, bool& found, float threshold = 0.4f);

            /**
             * Get the top match Candidate in the collection and the corresponding similarity score and match probability
             * for each probe Faceprint. Like identifyTopCandidate(), but runs search queries in parallel and improves throughput.
             *
             * @param[in] faceprints a vector of Faceprint to be identified.
             * @param[out] candidates the top match Candidate for each probe Faceprint.
             * @param[out] found set to true if a match is found for the probe Faceprint.
             * @param[in] threshold the similarity score threshold above which it is considered a match.
             * Higher thresholds may result in faster queries.
             * Refer to out <a href="https://docs.trueface.ai/ROC-Curves-d47d2730cf0a44afacb39aae0ed1b45a">ROC curves</a> when selecting a threshold.
             *
             * @return error code, see ErrorCode.
             */
            ErrorCode batchIdentifyTopCandidate(const std::vector<Faceprint>& faceprints, std::vector<Candidate>& candidates,
                    std::vector<bool>& found, float threshold = 0.4f);

            /**
             * Get a list of the top n match Candidate in the collection and their corresponding similarity scores and match probabilities.
             * Ideal for investigation mode.
             *
             * @param[in] faceprint the Faceprint to be identified.
             * @param[out] candidates a list of the top n matches (or fewer) Candidate, ordered by descending similarity score.
             * @param[out] found set to true if at leaset one match is found.
             * @param[in] numCandidates max number of Candidate identities to return.
             * @param[in] threshold the similarity score threshold above which it is considered a match.
             * Higher thresholds may result in faster queries.
             * Refer to out <a href="https://docs.trueface.ai/ROC-Curves-d47d2730cf0a44afacb39aae0ed1b45a">ROC curves</a> when selecting a threshold.
             *
             * @return error code, see ErrorCode.
             */
            ErrorCode identifyTopCandidates(const Faceprint& faceprint, std::vector<Candidate>& candidates, bool& found,
                            unsigned int numCandidates, float threshold = 0.4f);

            /**
             * Detect whether the face in the image is wearing a mask or not
             *
             * @param[in] faceBoxAndLandmarks FaceBoxAndLandmarks returned by detectFaces() or detectLargestFace().
             * @param[out] result The predicted MaskLabel for face image.
             * @param[out] maskScore The mask score for this image. This can be used for setting custom thresholds that work
             * better for the use case. By default, we use a mask score greater than 3.0 to determine that a mask was detected.
             *
             * @return error code, see ErrorCode.
             */
            ErrorCode detectMask(const FaceBoxAndLandmarks& faceBoxAndLandmarks, MaskLabel& result, float& maskScore);

            /**
             * Detect whether the face in the image is wearing any type of eye glasses or not
             *
             * @param[in] faceBoxAndLandmarks FaceBoxAndLandmarks returned by detectFaces() or detectLargestFace().
             * @param[out] result The predicted GlassesLabel for face image.
             * @param[out] glassesScore The glasses score for this image. This can be used for setting custom thresholds that work
             * better for the use case. By default, we use a glasses score greater than 0.0 to determine that glasses were detected.
             *
             * @return error code, see ErrorCode.
             */
            ErrorCode detectGlasses(const FaceBoxAndLandmarks& faceBoxAndLandmarks, GlassesLabel& result, float& glassesScore);

    private:
            class Impl;
            std::unique_ptr<Impl> pImpl;
    };

    class ObjectTracker {
        public:
            ObjectTracker(int numPoints, float alpha=0.5, float beta=0.005);
            Point<float> UpdateAndPredict(int pointIndex, Point<float> point);
        private:
            float alpha, beta;
            std::vector<Point<float>> points;
            std::vector<Point<float>> velocities;
            std::vector<bool> initialized;
    };
} // namespace truface
